@@ -55,3 +55,24 @@ test_that("calculate_d works for odd perm_order", {
 test_that("calculate_d works for identity", {
   expect_equal(calculate_d(1), 1)
 })
+
+test_that("get_structure_constants checks for proper argument", {
+  expect_silent(get_structure_constants(gips_perm("(1,2,3)", 3)))
+
+  expect_error(
+    get_structure_constants(3),
+    "You provided `perm` with `class\\(perm\\) == \\(numeric\\)`"
+  )
+
+  expect_error(
+    get_structure_constants(permutations::as.cycle("(1,2,3)")),
+    "You provided `perm` with `class\\(perm\\) == \\(permutation, cycle\\)`"
+  )
+})
+
+test_that("get_structure_constants can get a gips as perm", {
+  perm <- gips_perm(permutations::as.word(c(1, 2, 3, 5, 4)), 5)
+  expect_silent(constants1 <- get_structure_constants(perm))
+  expect_silent(constants2 <- get_structure_constants(gips(diag(5), 5, perm = perm)))
+  expect_equal(constants1, constants2)
+})
